import unittest
from InCli.SFAPI import restClient,CPQ,account,Sobjects,utils,query,jsonFile

class Test_Query(unittest.TestCase):
    def test_createCart(self):
        restClient.init('NOSDEV')

        acc = account.createAccount('Name:unaiTest2',recordTypeName='Consumer')

        self.assertTrue(Sobjects.checkId(acc['Id']))

        cartId = CPQ.createCart(accountF= acc['Id'], pricelistF='Name:WOO Price List',name="testa4",checkExists=True)

        self.assertTrue(Sobjects.checkId(cartId))

        promotions = CPQ.getCartPromotions_api(cartId)


        promotionId = query.queryField(" select Id from  vlocity_cmt__Promotion__c where vlocity_cmt__Code__c='PROMO_NOS_OFFER_016'")
      #  promos = []
      #  for promotion in promotions:
      #      promo = {
      #          "name":promotion['Name'],
      #          "code":promotion['vlocity_cmt__Code__c']
      #      }
      #      promos.append(promo)
      #      if promotion['vlocity_cmt__Code__c'] == 'PROMO_NOS_OFFER_016':
      #          promotionId = promotion['Id']

      #  utils.printFormated(promos)

        res = CPQ.postCartsPromoItems_api(cartId,promotionId)

        print()








from datetime import datetime
from logging import log
import logging
from . import objectUtil
import inspect,simplejson
import os,sys
from posixpath import dirname
import time
import subprocess


CBLACK = "\033[0;30m"
CRED = "\033[0;31m"
CGREEN = "\033[0;32m"
CBROWN = "\033[0;33m"
CBLUE = "\033[0;34m"
CPURPLE = "\033[0;35m"
CCYAN = "\033[0;36m"
CLIGHT_GRAY = "\033[0;37m"
CDARK_GRAY = "\033[1;30m"
CLIGHT_RED = "\033[1;31m"
CLIGHT_GREEN = "\033[1;32m"
CYELLOW = "\033[1;33m"
CLIGHT_BLUE = "\033[1;34m"
CLIGHT_PURPLE = "\033[1;35m"
CLIGHT_CYAN = "\033[1;36m"
CLIGHT_WHITE = "\033[1;37m"
CWHITE = '\33[37m'
CBOLD = "\033[1m"
CFAINT = "\033[2m"
CITALIC = "\033[3m"
CUNDERLINE = "\033[4m"
CBLINK = "\033[5m"
CNEGATIVE = "\033[7m"
CCROSSED = "\033[9m"
CEND = "\033[0m"

CBGBLUE = "\033[104m"

#https://stackoverflow.com/questions/287871/how-do-i-print-colored-text-to-the-terminal
#----------------------------------------------------------------------------------------------------
def stringifyUrlParameters(parameters,exclude,initSeparator='?'):
  isFirst = True

  ret = ''
  for name in parameters:
    if name in exclude:
      continue
    if(parameters[name] == None):
      continue
    separator = initSeparator if isFirst == True else '&'
    value = parameters[name]
    if isinstance(value, str) == False:
      value = simplejson.dumps(value)
    paramStr = f'{separator}{name}={value}'
    ret = ret + paramStr
    isFirst = False

  return ret

def executeCommand(command):
    is_windows = sys.platform.startswith('win')
    if is_windows == True:
        list_files = subprocess.run(command,capture_output=True,text=True,shell=True)
    else:
        list_files = subprocess.run(command,capture_output=True,text=True)

    return list_files

def execute_force_org_display(userName_or_orgAlias):
    def getValue(field):
        if field in line:
            chunks = line.split(field)
            val = chunks[1].strip()
            obj[field]=val
    output = executeCommand(["sfdx","force:org:display","-u", userName_or_orgAlias])
    if output.stdout == '':
        return False,None,output

    lines = output.stdout.splitlines()

    obj = {}
    for line in lines:
        getValue('Access Token')
        getValue('Connected Status')
        getValue('Instance Url')

    return True,obj,output

def executeCommandParse(command):
    output = executeCommand(command)

    if output.stdout == '':
        return False,output

    obj = {}

    sl = output.stdout.splitlines()
    namesline = sl[1]
    lenLine = sl[2]
    lenLines = lenLine.split(' ')
    ll =[0]
    ll = ll+[len(lenline)+2 for lenline in lenLines if lenline!='']
    ll[-1] = ll[-1]-2

    pos = []
    position = 0
    for l in ll:
        position = position + l
        pos.append(position)


    names = []
    for x in range(len(pos)-1):
        name = namesline[pos[x]:pos[x+1]].strip()
      #  print(namesline[pos[x]:pos[x+1]])
        if name == '':
            name = f"{x}"
        names.append(name)

    objs =[]
    for y in range(3,len(sl)-1):
        obj = {}
        objs.append(obj)
        for x in range(len(pos)-1):
            value = sl[y][pos[x]:pos[x+1]].strip()
            obj[names[x]] = value

    return True,objs
#----------------------------------------------------------------------------------------------------
def extendedField(field):
    """
    An extended field is a field defined as "field:value"
    returns a json containing {'field':field,'value':value}
    If the parameter is just a value, the field is set to Id  'value' --> 'Id:value'
    """
    if field is None:
        return None

    sp = field.split(':')
    if len(sp) == 1:
        return {
            'field':'Id',
            'value':sp[0]
        }
    return {
        'field':sp[0],
        'value':sp[1]
    }
#----------------------------------------------------------------------------------------------------
class InCliError(Exception):
    pass

def raiseException(errorCode,error,module=None,other=None):
    aa = inspect.stack()[1]
    error = {
        'errorCode':errorCode,
        'error':error,
        'module':f"{aa[1].split('/')[-1]}  {aa[3]}  {aa[2]}",
    }
    if other is not None:
        error['other'] = other
    
    error['json'] = message = simplejson.dumps(error, indent=2)

    raise InCliError(error)

def printException(exception):
    print(type(exception).__name__)
    if type(exception) == InCliError:
        _str = exception.args[0]['json']
    else:
        error = {
            "errorCode":type(exception).__name__,
            "error":str(exception)
        }
        _str = simplejson.dumps(error, indent=2)
    logging.error(_str)

#----------------------------------------------------------------------------------------------------

def throwOrLog(message,raiseEx=True,dumps=False):
    """
    if raiseEx is True, it raises an error. IF not logs the error. 
    - message: the error messages. Can be a dictionary
    - raiseEx: boolean
    """
    if dumps == True:
        message = simplejson.dumps(message, indent=2)

    if raiseEx == True:
        logging.error(message)
        raise ValueError(message)
    else:
        logging.info(message)
        return None
#----------------------------------------------------------------------------------------------------
def printJSON(obj,delNull=False):
    if delNull == True:
        obj = deleteNulls(obj)
    d = simplejson.dumps(obj, indent=2)
    logging.debug(d)
    print()

def deleteNulls(obj,systemFields=True,nulls=True,delAttributes=True):
    systemFields= ['OwnerId','IsDeleted','CreatedDate','CreatedById','LastModifiedDate','LastModifiedById','LastViewedDate','LastReferencedDate','SystemModstamp','attributes']
  #  if systemFields == False:
  #      systemFields = True
  #      systemFields = ['attributes']

    if systemFields == True and nulls == True:
        return obj
    if 'records' in obj:
        obj = obj['records']
    if type(obj) is list:
        res = []
        for ob in obj:
            res.append(deleteNulls(ob,systemFields,nulls))
        return res

    res = {}
    for key in obj.keys():
        if systemFields == False:
            if key in systemFields:
                continue
        if delAttributes == True:
            if key in ['attributes']:
                continue
        if type(obj[key]) is dict:
            res[key] = deleteNulls(obj[key],systemFields,nulls)
        else:
            if obj[key] is None and nulls is False:
                continue
            res[key] = obj[key]
    return res

def printFormated(records,fieldsString=None,rename=None,exclude=None):
    """
    - fieldsString: format field%name:field2%name2:field3
        field is the field in the obj. Name is how it will be printed (the short version). 
        name3 is equivalent to field3-field3
    - rename : field%name:field2%name2
    """
    if records is None or len(records) == 0:
        print('No records to print')
        return

    print()
    if 'records' in records:
        records = records['records']

    if fieldsString == None:
        fieldsString = ":".join(records[0].keys())
    fields = fieldsString.split(':')
    if exclude != None:
        fields2 = [field for field in fields if field not in exclude.split(':')]
        fields = fields2

    _renames = {}
    if rename != None:
        renames = rename.split(':')
        for ren in renames:
            chunks = ren.split('%')
            _renames[chunks[0]] = chunks[1]
            print( CFAINT +  f"Column {chunks[0]} as {chunks[1]}" +CEND)
        print()
    
    _fields = {}
    for field in fields:
        v = {}
        v['obj_name'] = field
        v['print_name'] = _renames[field] if field in _renames else field
        v['size'] = len(v['print_name'])
        _fields[field] = v

    for record in records:
        for key in _fields.keys():
            size = 0
            value = objectUtil.getField(record,key,'.')
            if value is not None:
                size = len(str(value))
                if size > _fields[key]['size']:
                    _fields[key]['size'] = size

    spacing = 2    
    line = ''
    for key in _fields.keys():
        if key == "__color__":
            continue
        field = _fields[key]        
        value = field['print_name']

        line =  f"{line}{value:{int(field['size']+spacing)}}" 
    print(CUNDERLINE+line+CEND)

    for record in records:
        line = ''
        for key in _fields.keys():
            field = _fields[key]    

            if key == "__color__":
                continue

            val = objectUtil.getField(record,key,'.')
            val = '' if val is None else str(val) 
            padding = int(field['size']+spacing)
 #           print(f"{padding}  {field['size']}")
            if _isInt(val):
                line = f"{line}{int(val):>{padding-1}} "
            elif _isFloat(val):
                line = f"{line}{float(val):>{padding-1}.2f} "
            else:
                line = f"{line}{val:{padding}}"
        if '__color__' in record:
            if record['__color__'] != '':
                line = record['__color__'] + line + CEND

        print(line)

def _isInt(s):
    try:
        int(s)
        return True
    except ValueError:
        return False

def _isFloat(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def datetimeString_now(format = '%Y-%m-%dT%H:%M:%S'):
    ts = time.time()
    dt = datetime.fromtimestamp(ts).strftime(format)
    return dt

def datetimeString_to_timestamp(str,format=None):
    if str == None: 
        return None
    if format == None:
        format = '%Y-%m-%dT%H:%M:%S.%f%z'
    #str = '2021-08-31T22:23:24.000+0000'   '%Y-%m-%dT%H:%M:%S.%f%z'
    dt = datetime.strptime(str,format)
    ts = datetime.timestamp(dt) * 1000
    
    return int(ts)

def today_date(format="%Y-%m-%d"):
    today = datetime.today()
    return today.strftime(format)  

def future_date(years=None,days=None,format="%Y-%m-%d"):
    today = datetime.today()
    if years!=None:
        future = today.replace(year=today.year + years)
    if days!=None:
        future = today.replace(day=today.day + days)
    return future.strftime(format)  

def callerPythonFileFolder():
    caller_filepath = inspect.currentframe().f_back.f_code.co_filename
    caller_filefolder = os.path.dirname(caller_filepath)
    if caller_filefolder.endswith('SFAPI'):
        caller_filefolder = dirname(caller_filefolder)
    return caller_filefolder


def getEntryDirectory():
    return sys.path[0]

def currentFilePath():
    frame = inspect.stack()[1]
    filename = frame[0].f_code.co_filename
    dirname = os.path.dirname(filename)
    return dirname

def getFieldc(var,field,onlyOne=True):
    return Id(var,field=field,onlyOne=True)

def getRecord(obj):

    if 'records' in obj:
        return getRecord(obj['records'])
    if 'result' in obj:
        return getRecord(obj['result'])
    if 'fields' in obj:
        return getRecord(obj['fields'])
    return obj

def Id(var):
    return selectField(var,field='Id',onlyOne=True)

def selectField(var,field='Id',onlyOne=True):
    if type(var) is str:
        return var
    if (type(var) is dict):
        if field in var:
            if 'value' in var[field]:
                return var[field]['value']
            return var[field]

        if 'records' in var:
            return selectField(var['records'],field,onlyOne)
        if 'result' in var:
            return selectField(var['result'],field,onlyOne)
        if 'fields' in var:
            return selectField(var['fields'],field,onlyOne)
    if type(var) == list:
        if onlyOne == True:
            if len(var)>1:
                logging.warn(f"There is more than 1 element in the list {len(var)}, returning Id from first one")
            return selectField(var[0],field,onlyOne)
        else:
            _ids = []
            i=0
            while i<len(var):
                _ids.append(selectField(var[i],field,onlyOne))
                i = i + 1
            return _ids

def Obj(var,onlyOne=True):
    if (type(var) is dict):
        if 'records' in var:
            return Obj(var['records'],onlyOne)
        else :
            return var

    if type(var) == list:
        if onlyOne == True:
            if len(var)>1:
                logging.warn(f"There is more than 1 element in the list {len(var)}, returning Id from first one")
            return Obj(var[0],onlyOne)
        else:
            _objs = []
            i=0
            while i<len(var):
                _objs.append(Obj(var[i],onlyOne))
                i = i + 1
            return _objs


def computeTimes(call,name,times,_type=None):
    global _counter

    debugTimes = {}

    time = call['elapsedTime']
    sec = time.microseconds + time.seconds * 1000000 
    times[name]=sec

    debugTimes['ElapsedCall'] = sec

    if 'debugIP' == _type and 'IPResult' in call:

        if 'debugLog' in call['IPResult']:
            for key in call['IPResult']:
                if key.endswith('Debug'):
                    debugTimes[f'{key}_ElapsedTime'] = call['IPResult'][key]['deltaTime'] * 1000
                    debugTimes[f'{key}_ElapsedTimeCPU'] = call['IPResult'][key]['ElapsedTimeCPU'] * 1000

        if 'elapsedTimeActual' in call['IPResult']:
            debugTimes['elapsedTimeActual'] = call['IPResult']['elapsedTimeActual'] * 1000
        if 'elapsedTimeCPU' in call['IPResult']:
            debugTimes['elapsedTimeCPU'] = call['IPResult']['elapsedTimeCPU'] * 1000
            #print(f" {_counter} , {sec}  , {elapsedTimeActual}  , {sec-timeActual} ,{apigeeTime}")
            #_counter = _counter + 1
            #return
        print(debugTimes)

    print (sec)
    return debugTimes

def repeatingSequence(objs,field):

    repeatingSequences = []
    lastX = 0
    for x,val in enumerate(objs):
        if x < lastX:
            continue
        firstdelta = None
        lastDelta = 0
        positions = []

        for y in range(x+1,len(objs)):

            if field not in objs[y]:
                _o = objs[y]
                print()
            if objs[y][field] == objs[x][field]:
                line = objs[y][field]
                if firstdelta == None:
                    firstdelta = y-x
                    lastDelta = firstdelta
                else:
                    lastDelta = y - lastX
                
                seq = True
                if lastDelta == firstdelta:
                    for z in range(0,firstdelta):
                        if y+z >= len(objs):
                            seq=False
                            break
                        if objs[y+z][field] != objs[x+z][field]:
                            seq = False
                            break
                    if seq == True:
                        if len(positions) == 0:
                            positions.append(y-firstdelta)
                        positions.append(y)
                else:
                    break
                if seq == True:
                    lastX = y

        if len(positions) > 0:
            repeatingSequences.append(positions)
    #        print(f"{positions}   {firstdelta}" )
    #        for p in positions:
    #            for m in range(0,firstdelta):

     #               print(f"{p+m}  {objs[p+m][field]}")
     #       print()

    return repeatingSequences
    print(repeatingSequences)


from . import Sobjects,query,utils
import logging

#if the account does not exist, accountX will be used as the account name. 
def createAccount_Id(accountF,fields=None,recordTypeName=None,checkExists=True):

    ef = utils.extendedField(accountF)
    if checkExists == True:
        accountId = query.queryField(f" select Id from Account where {ef['field']}='{ef['value']}' ")
        if accountId != None:
            logging.info(f'Returning existing Account <{accountId}> ')
            return accountId
            
    acc={
        'Name': ef['value'],
        'vlocity_cmt__Active__c':'Yes',
        'vlocity_cmt__Status__c':'Active'
    }
    if fields!= None:
        acc.update(fields)

    if recordTypeName!=None:
        acc['RecordTypeId'] = query.queryField(f"select Id from RecordType where SobjectType='Account' and Name = '{recordTypeName}'")

    call = Sobjects.create('Account',acc)

    accountId = call['id']
    logging.info(f"Account {ef['value']} Created Account <{accountId}> ")

    return accountId

def createAccount(accountF,fields=None,recordTypeName=None,checkExists=True):
    accountId = createAccount_Id(accountF,fields=fields,recordTypeName=recordTypeName,checkExists=checkExists)
    account = Sobjects.get(accountId)

    return account['records'][0]

def getAccount(accountF):
    ef = utils.extendedField(accountF)
    return query.query(f" select fields(all) from Account where {ef['field']}='{ef['value']}' limit 100")

def deleteAccount(accountF):
    ef = utils.extendedField(accountF)

    ids = query.queryFieldList(f" select Id from Account where {ef['field']}='{ef['value']}' ")
    return Sobjects.delete('Account',ids)

def deleteAccountOrders(accountId):
    ids = query.queryField(f" select Id from Order where  AccountId = '{accountId}' ")
    return Sobjects.delete('Account',ids)



from InCli.SFAPI import restClient ,utils,file,jsonFile,query,thread

import unittest,logging,os,shutil

class Test_Overrides(unittest.TestCase):
    def call_ServicesData(self):
        action = '/services/data'
        call = restClient.callAPI(action)
        self.assertTrue(len(call)>0)  

    def test_configFile(self):
        restClient.setLoggingLevel(logging.INFO)
        try:
            restClient.setConfigFile('sss')
        except Exception as e:
            utils.printException(e)
            self.assertTrue(e.args[0]['errorCode'] == 'NO_CONFIG')

        restClient.setConfigFile('/Users/uormaechea/Documents/Dev/python/Industries/config/ConnectionsParams.json')

        try:
            restClient.initWithConfig('XXXX')
        except Exception as e:
            utils.printException(e)
            self.assertTrue(e.args[0]['errorCode'] == 'NO_ORG')

        try:
            restClient.initWithConfig('NOSDEV')
            self.assertTrue(restClient._currentConnectionName=='NOSDEV')

            folder = restClient.debugFolder()

            st = '2222222'
            file.write(f"{folder}test.txt",st)
            st2 = file.read(f"{folder}test.txt")
            self.assertTrue(st==st2)

            
        except Exception as e:
            utils.printException(e)
            self.assertTrue(1==2)
        print()

    def test_debug_action(self):
     #   restClient.setConfigFile('/Users/uormaechea/Documents/Dev/python/Industries/config/ConnectionsParams.json')
        restClient.init("DEVNOSCAT2")   
        action = '/services/data'

        #Test  callAPI_debug 
        call = restClient.callAPI(action)
        self.assertTrue(len(call)>0)    
        restClient.callSave("testFile")
        lc = restClient.lastThreadCall()

        self.assertTrue('responseFilePath' in lc)
        fileContent = jsonFile.read(lc['responseFilePath'])

        self.assertEqual(call,fileContent)

        q = query.query(" select fields(all) from Account limit 1")
        print(q)

 
    def test_debug_guest(self):
        restClient.setConfigFile('/Users/uormaechea/Documents/Dev/python/Industries/config/ConnectionsParams.json')

        restClient.initWithConfig('DEVNOSCAT2_GUEST')
        action = '/services/data'
        call = restClient.callAPI(action)
        self.assertTrue(len(call)>0)    

        try:
            q = query.query(" select fields(all) from Account limit 1")
        except Exception as e:
            utils.printException(e)
            self.assertTrue(e.args[0]['errorCode'] ==  'HTTPs Error: 401')

    def test_debug_withToken(self):
        restClient.setLoggingLevel(logging.DEBUG)
        restClient.init('DEVNOSCAT2')
        
        action = '/services/data'

        call = restClient.callAPI(action)
        self.assertTrue(len(call)>0)  
        
        url = restClient._initializedConnections[0]['instance_url']
        token = restClient._initializedConnections[0]['access_token']
    
        restClient.initWithToken('test',url,token=token)
        self.assertTrue(restClient._currentConnectionName == 'test')
        call = restClient.callAPI(action)
        self.assertTrue(len(call)>0)  

        q = query.query(" select fields(all) from Order limit 1")
        print(q)

    def test_debug_threaded(self):
        restClient.init("DEVNOSCAT2") 

        a = [1]*100

        allTimes = []
        def doWork(a):
            action = '/services/data'
            call = restClient.callAPI(action)
            times= {
                "elapsed":restClient.getlastThreadCallTime()
            } 
            allTimes.append(times)

        thread.processList(doWork,a,10)

        utils.printFormated(allTimes)

    def test_login_sfdx(self):
        try:
            outputs = utils.executeCommandParse(["sfdxXXXX","auth:web:login","-r", "https://nos--devnoscat2.sandbox.my.salesforce.com"])
        except Exception as e:
            self.assertTrue(e.strerror ==  'No such file or directory')
            utils.printException(e)

        try:
            restClient.init("xxx.xxx@nos.pt")

        except Exception as e:
            self.assertTrue(e.args[0]['errorCode'] ==  'ConnectionError')
            utils.printException(e)

        try:            
            restClient.init("uormaechea.devnoscat2@nos.pt")
            q = query.query(" select fields(all) from Order limit 1")
            lc = restClient.lastThreadCall()
            self.assertEqual(lc['status_code'],200)
            print(q)

        except Exception as e:
            self.assertEqual('This','Should not happen')

    def test_congifData(self):        
        return
        dir = os.path.abspath("tmp")
        try:
            os.mkdir(dir)
        except Exception as e:
            self.assertTrue(e.strerror=='File exists')
        os.chdir(dir)
        current = os.getcwd()

        config = restClient.loadConfigData()

        self.assertTrue('orgs' in config)
        self.assertTrue(file.exists(restClient._configDataName))

        shutil.rmtree(current)


    def test_saveOrg(self):
        restClient.init("DEVNOSCAT2") 

        url = restClient._initializedConnections[0]['instance_url']
        token = restClient._initializedConnections[0]['access_token']

        restClient.saveOrg_inConfigFile('test1',url,token)
        restClient.initWithConfig('test1')
        self.assertTrue(restClient._currentConnectionName=='test1')
        self.call_ServicesData()

        restClient.saveOrg_inConfigFile('testGuest',url)
        restClient.initWithConfig('testGuest')
        self.assertTrue(restClient._currentConnectionName=='testGuest')
        self.call_ServicesData()
        try:
            q = query.query(" select fields(all) from Order limit 1")
        except Exception as e:
            self.assertEqual(e.args[0]['errorCode'],'HTTPs Error: 401')
            utils.printException(e)

    def test_saveDeleteOrg(self):
        restClient.saveOrg_inConfigFile('test1','xxx','yyyy')

        cd = restClient.loadConfigData()
        org = [i for i in cd['orgs'] if (i['name'] == 'test1')][0]
        self.assertTrue(org['name'] == 'test1')
        restClient.deleteOrg_inConfigFile('test1')
        cd = restClient.loadConfigData()
        org = [i for i in cd['orgs'] if (i['name'] == 'test1')]
        self.assertTrue(len(org)==0)
        print()




import unittest
from InCli.SFAPI import DicPath

class Test_DicPatj(unittest.TestCase):
    def test_simple(self):
        obj = {
            "a":1,
            "b":2,
            "c":3
        }
        res = DicPath.find(obj,"c")
        self.assertTrue(res==obj['c'])

        res = DicPath.find(obj,"c")

        print()


    def test_simple2(self):
        obj = {
            "a":1,
            "b":2,
            "c":{
                "a":1,
                "b":2,
                "c":3
            }
        }
        res = DicPath.find(obj,"c.c")
        self.assertTrue(res==obj['c']['c'])

        print()

        




from InCli.SFAPI import utils
import sys, os
#python3 -m unittest

import unittest
from InCli import InCli
#from InCli.SFAPI import restClient
import traceback
class Test_Main(unittest.TestCase):

    def test_q(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 1"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_q_System(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 1","-system"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_q_nulls(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 1","-null"])       
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2) 
    def test_q_all(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 1","-all"])   
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_q_fields_all(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select fields(all) from Order limit 10","-fields","AccountId"])  
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_q_fields_all(self):
        try:
            InCli._main(["-u","NOSDEV","-q", "select AccountId,Pricebook2Id,OrderNumber,TotalContractCost__c,State__c from Order limit 50","-fields","all"])  
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
    def test_o(self):
        try:
            InCli._main(["-u","uormaechea.devnoscat2@nos.pt","-o"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_o_name(self):
        try:
            InCli._main(["-u","uormaechea.devnoscat2@nos.pt","-o","-name","order"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_o_like(self):
        try:
            InCli._main(["-u","uormaechea.devnoscat2@nos.pt","-o","-like","XOM"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_h(self):
        try:
            InCli._main(["-h"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_cc(self):
        try:
            InCli._main(["-u","NOSQSM","-cc"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
            
    def test_cc_list(self):
        try:
            InCli._main(["-u","NOSDEV","-cc","-list"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_cc_code(self):
        try:
            InCli._main(["-u","NOSDEV","-cc","-code","TESTERRORS"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())

    def test_cc_account(self):
        try:
            InCli._main(["-u","NOSDEV","-cc","-account","name:unaiTest4","-code","TESTERRORS"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())

    def test_logs(self):
        InCli._main(["-u","NOSDEV","-logs"])

    def test_logs_user(self):
        InCli._main(["-u","NOSDEV","-logs","-loguser","Alias:ana.r"])

    def test_logs_query(self):
        InCli._main(["-u","NOSDEV","-logs","-where","Operation='Batch Apex'","-last","10"])

    def test_logs_userDefault(self):
        try:
            InCli._main(["-default:set","loguser","Alias:ana.r"])
            InCli._main(["-u","NOSDEV","-logs"])
            InCli._main(["-default:del","loguser"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs_userWrong(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","-loguser","Alias:xxxx"])
        except Exception as e:
            self.assertTrue(e.args[0]['error']=='User with field Alias = xxxx does not exist in the User Object.')
            utils.printException(e)

    def test_logs_limit(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","-limit","2"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_logs(self):
        try:
            InCli._main(["-u","NOSDEV","-logs"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)
   
    def test_log_ID_2(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","07L3O00000DgMVOUA3"])
        except Exception as e:
            print(e)
            print(traceback.format_exc())
    def test_log_ID(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","07L3O00000DgMguUAF"])
        except Exception as e:
            print(e)
            print(traceback.format_exc())
    def test_log_file(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","07L3O00000DfytZUAR","-level","0"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_log_last(self):
        try:
            InCli._main(["-u","NOSDEV","-logs","-last","10"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_default(self):
        try:
            InCli._main(["-default:set","u"])
            InCli._main(["-default:set","u","NOSDEV"])
            InCli._main(["-default:get","u"])        
            InCli._main(["-logs","-last","1"])
            InCli._main(["-default:del","u"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_d(self):
        try:
            InCli._main(["-u","NOSDEV","-d"])
            InCli._main(["-u","NOSDEV","-d","Order"])
            InCli._main(["-u","NOSDEV","-d","Order:Status"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)

    def test_l(self):
        try:
            InCli._main(["-u","NOSDEV","-l"])
            InCli._main(["-u","DEVNOSCAT2","-l"])
        except Exception as e:
            utils.printException(e)
            print(traceback.format_exc())
            self.assertTrue(1==2)


from . import restClient,query,digitalCommerceUtil, utils
import simplejson,logging,inspect

#------------------------------------------------------------------------

def stringify(parameters,exclude):
  isFirst = True

  ret = ''
  for name in parameters:
    if name in exclude:
      continue
    if(parameters[name] == None):
      continue
    separator = '?' if isFirst == True else '&'
    value = parameters[name]
    if isinstance(value, str) == False:
      value = simplejson.dumps(value)
    paramStr = f'{separator}{name}={value}'
    ret = ret + paramStr
    isFirst = False

  return ret

#------------------------------------------------------------------------

def checkBasketError():

  lc = restClient.lastCall()
  if lc['error'] is not None:
      utils.raiseException(lc['errorCode'],lc['error'],lc['action'])
  if lc['response']['error']!='OK':
      utils.raiseException(lc['response']['errorCode'],lc['response']['error'],other=lc['action'])
  if 'success' in lc['response'] and lc['response']['success'] == False:
      utils.raiseException(lc['response']['errorCode'],f"{inspect.stack()[1].function}: success:{lc['response']['success']}  isBasketValid:{lc['response']['isBasketValid']}")
  if 'offerDetails' in lc['response']['result']:
      utils.raiseException(f"CreateBasket {lc['response']['result']['offerDetails']['StatusCode']}",lc['response']['result']['offerDetails']['messages'][0])

  #if lc['response']['result']['error'] !='OK':
  #  print()

def checkOfferError():

  lc = restClient.lastCall()
  if lc['error'] is not None:
      utils.raiseException(lc['errorCode'],lc['error'],lc['action'])
  if lc['response']['error']!='OK':
      utils.raiseException(lc['response']['errorCode'],lc['response']['error'],other=lc['action'])
  if 'success' in lc['response'] and lc['response']['success'] == False:
      utils.raiseException(lc['response']['errorCode'],f"{inspect.stack()[1].function}: success:{lc['response']['success']}  isBasketValid:{lc['response']['isBasketValid']}")



#------------------------------------------------------------------------

def getResult(obj):

  if 'result' in obj:
    obj = obj['result']  
  return obj

#------------------------------------------------------------------------

def getCatalogs(catalogueCode=None):
    where = f" where vlocity_cmt__CatalogCode__c = '{catalogueCode}' " if catalogueCode !=None else ''
    call = query.query(f'select Fields(ALL) from vlocity_cmt__Catalog__c {where} limit 100')
    return call['records']

#------------------------------------------------------------------------

def getCatalogProducts(catalogId):
    call = query.query(f"select Fields(ALL) from vlocity_cmt__CatalogProductRelationship__c where vlocity_cmt__CatalogId__c='{catalogId}' limit 100")  
    return call

#------------------------------------------------------------------------

def getCatalogbyField(value,field='Name'):
    call = query.query(f"select Fields(ALL) from vlocity_cmt__Catalog__c where {field}='{value}' limit 100")    
    return call

#------------------------------------------------------------------------

def getOfferByCatalogue(catalogcode,contains=None,contextkey=None,context=None,forceinvalidatecache=None,isloggedin=None,pagesize=None,offset=None,sortby=None,excludeoverridedefinition=None,offerList=None):
  paramStr = stringify(locals(),exclude=['catalogcode'])
  action = f'/services/apexrest/{restClient.getNamespace()}/v3/catalogs/{catalogcode}/offers{paramStr}'
  call = restClient.callAPI(action)
  checkOfferError()

  if 'offers' in call:
    return call['offers']

  return []

#------------------------------------------------------------------------

def getOfferDetails(catalogcode,offercode,contextkey=None,context=None,forceinvalidatecache=None,isloggedin=None):
  paramStr = stringify(locals(),exclude=['catalogcode','offercode'])
  action = f'/services/apexrest/{restClient.getNamespace()}/v3/catalogs/{catalogcode}/offers/{offercode}{paramStr}'
  call = restClient.callAPI(action)
  checkOfferError()


  lc = restClient.lastCall()
  log = {
      'method':'getOfferDetails',
      'action':f"{lc['url']}{lc['action']}",
      'catalog':catalogcode,
      'offer':offercode,
      'contextKey':call['contextKey'],
      'error':call['error'],
      'deltaTime':lc['deltaTime'],
      'elapsedTime':lc['elapsedTime']
  }
  lc['log'] = log
  return call
#------------------------------------------------------------------------

def createBasket(catalogcode,offer,contextKey=None,context=None,forceinvalidatecache=None,isloggedin=None,includeAttachment=None,returnBasket=None,validatebasket=None):
  paramStr = stringify(locals(),exclude=['catalogcode','offer'])
  action = f'/services/apexrest/{restClient.getNamespace()}/v3/catalogs/{catalogcode}/basket{paramStr}'
  restClient.glog().debug(action)

  body = {
    "basketAction": "AddWithNoConfig",
    "offer":offer
  }

  response = restClient.callAPI(action, method="post", data=body)
  checkBasketError()

  lc = restClient.lastCall()
  log = {
      'method':'AddWithNoConfig',
      'action':f"{lc['url']}{lc['action']}",
      'catalog':catalogcode,
      'offer':offer,
      'contextKey':response['cartContextKey'],
      'error':response['error'],
      'deltaTime':lc['deltaTime'],
      'elapsedTime':lc['elapsedTime']              
  }
  lc['log'] = log

  return response

#------------------------------------------------------------------------

def createBasketAfterConfig(catalogcode,offerDetails,contextKey=None,context=None,forceinvalidatecache=None,isloggedin=None,includeAttachment=None,returnBasket=None,validatebasket=None):
  paramStr = stringify(locals(),exclude=['catalogcode','offerDetails'])
  action = f'/services/apexrest/{restClient.getNamespace()}/v3/catalogs/{catalogcode}/basket{paramStr}'
  restClient.glog().debug(action)

  body = {
    "basketAction": "AddAfterConfig",
    "productConfig": getResult(offerDetails)
  }

  restClient.writeFile('body12xx',body)
  response = restClient.callAPI(action, method="post", data=body)
  checkBasketError()


  lc = restClient.lastCall()
  log = {
      'method':'AddAfterConfig',
      'action':f"{lc['url']}{lc['action']}",
      'catalog':catalogcode,
      'offer':getOfferCode( offerDetails['result']['offerDetails']['offer']),
      'contextKey':response['cartContextKey'] if 'cartContextKey' in response else "",
      'error':response['error'],
      'deltaTime':lc['deltaTime'],
      'elapsedTime':lc['elapsedTime']              
  }
  lc['log'] = log

  return response

#------------------------------------------------------------------------

def executeActions(basketAction):
  action = basketAction['link']
  body = basketAction['params']
  method="post"

  if basketAction['method'] == 'updateBasketAction':
    method = 'put'

  response = restClient.callAPI(action, method=method, data=body)
  response = checkBasketError(response)
  return response

#------------------------------------------------------------------------

def addChildBasket(basket,path):
  basketAction = digitalCommerceUtil.getAction(basket,path,method='addChildBasketAction')
  basket2 = executeActions(basketAction)

  return basket2

#------------------------------------------------------------------------

def deleteFromBasket(basket,path):
  basketAction = digitalCommerceUtil.getAction(basket,path,method='deleteFromBasketAction')
  basket2 = executeActions(basketAction)

  return basket2

#------------------------------------------------------------------------

def setError(error,errorCode):
    restClient.glog().error(f"Error {error}  {errorCode}")
    restClient.setLastCallError( error,errorCode)



#------------------------------------------------------------------------

def updateBasketQuantity(basket,path,value):
  basketAction = digitalCommerceUtil.getAction(basket,path,method='updateBasketAction')
  basketAction['params']['Quantity'] = str(value)
  basket2 = executeActions(basketAction)
  if len(basket2['result']['messages'])==0:
    print('ok')
  else:
    print("Error")
  return checkBasketError(basket2)

#------------------------------------------------------------------------

#{"field":value}
def updateBasketFields(basket,path,fieldsToBeUpdated):
  basketAction = digitalCommerceUtil.getAction(basket,path,method='updateBasketAction')
  basketAction['params']['fieldsToBeUpdated'] = fieldsToBeUpdated
  basket2 = executeActions(basketAction)

  return basket2


#  updates = [
#    {
#      "Category":'ACAT_EQUIPMENT',
#      "Attribute":'ATT_SERIAL_NUMBER',
#      "value":""
#    }
#  ]

#------------------------------------------------------------------------

def updateBasketAttributes(basket,path,updates):
  attributes = digitalCommerceUtil.getBasketProductAttributes(basket,path)

  for update in updates:
    for cat in attributes['records']:
      if cat['Code__c'] == update['Category']:
        for attribute in cat['productAttributes']['records']:
          if attribute['code'] == update['Attribute']:
            attribute['userValues'] = update['value']

  basketAction = digitalCommerceUtil.getAction(basket,path,method='updateBasketAction')
  basketAction['params']['attributeCategories'] = attributes
  basket2 = executeActions(basketAction)

  return basket2

#------------------------------------------------------------------------

def createCart(accountId,catalogCode,cartContextKey,createAsset=False,assetReferenceKey=None):
  if type(cartContextKey) is dict:
    cartContextKey = cartContextKey

  data = {
    "accountId": accountId,
    "catalogCode": catalogCode,
    "cartContextKey": cartContextKey
  }
  action = f'/services/apexrest/{restClient.getNamespace()}/v3/carts?price=false&validate=false'

  if assetReferenceKey != None:
    data['assetReferenceKey'] = assetReferenceKey
    context = 'context={"accountId":"' + accountId + '"}'
    action = f'/services/apexrest/{restClient.getNamespace()}/v3/carts?{context}&isloggedin=true&price=false&validate=false'
   # action = f'/services/apexrest/{client.getNamespace()}/v3/carts?{context}&isloggedin=true'

  if createAsset:
      action = f'{action}&createAsset=true'

  #print(data)
  call = restClient.callAPI(action, method="post", data=data)
  checkBasketError()
  return call

#------------------------------------------------------------------------

def printPricing(offerDetails):
  offer = offerDetails['result']['offerDetails']['offer']
  if 'priceResult' in offer:
    if isinstance(offer['priceResult'],list):
      for priceResult in offer['priceResult']:
        if 'DisplayText__c' in priceResult:
          print(f"{offer['Name']}  {priceResult['DisplayText__c']}")
    else:   #something wrong such as no valid price found.
      print(f"{offer['Name']}  {offer['priceResult']}")

  else:
    print('No price result')
  if 'childProducts' in offer:
    for childProduct in offer['childProducts']:
      if 'priceResult' in childProduct:
        for priceResult in childProduct['priceResult']:
            print(f"{childProduct['Name']}  {priceResult['DisplayText__c']}")
      else:
          print(f"{childProduct['Name']}  <No pricing>")

#------------------------------------------------------------------------

def getOfferCode(offer):
  if offer['offerType'] == 'Product':
      code = offer['ProductCode']
  else:
      code = offer['vlocity_cmt__Code__c']
    
  return code


import logging, copy
from . import Sobjects, objectUtil, priceBook, query, utils,restClient
from datetime import date
import simplejson 


_carts = []
def _get(Id):
    for cart in _carts:
        if cart['Id'] == Id:
            return cart
def _set(Id,name,pricelistId,accountId):
    global _carts
    _carts.append({
        'Id':Id,
        'name':name,
        'plId':pricelistId,
        'accId':accountId
    })
#------------------------------------------------------------------------

def stringify(parameters,exclude,initSeparator='?'):
  isFirst = True

  ret = ''
  for name in parameters:
    if name in exclude:
      continue
    if(parameters[name] == None):
      continue
    separator = initSeparator if isFirst == True else '&'
    value = parameters[name]
    if isinstance(value, str) == False:
      value = simplejson.dumps(value)
    paramStr = f'{separator}{name}={value}'
    ret = ret + paramStr
    isFirst = False

  return ret


#------------------------------------------------------------------------
def throw(message):
    msg = simplejson.dumps(message, indent=4)

    logging.error(msg)
    raise ValueError(msg)


#------------------------------------------------------------------------

def checkError():
    lc = restClient.lastCall()
    if lc['error'] is not None:
      utils.raiseException(lc['errorCode'],lc['error'],lc['action'])
 #   response = lastcall['response']
 #   if 'messages' in response:
 #       if len(response['messages']) > 0:
 #           msg = response['messages'][0]
 #           if 'code' in msg:
 #               if msg['code'] == '208' or msg['code'] == '101':
 #                   return False
#
#                if msg['severity'] == 'ERROR':
#                    return throw(msg)

 #   elif lastcall['status_code'] == 600:
 #       return throw(response)

    return False

#------------------------------------------------------------------------
def createQuote(opportunityid,pricelistName,name, recordTypeId,expirationDate = "2023-01-17" ,inputFields = None,fields='Id,Name,EffectiveDate'):
    global _carts

    Body = {
    "methodName": "createCart",
    "objectType": 'Quote',
    "inputFields": [
        {"opportunityid": opportunityid},
        {"Name": name},
        {"ExpirationDate": expirationDate},
        {"RecordTypeId": recordTypeId},
        {"pricelistName": pricelistName}
        ],
    "subaction":'createQuote',
    "fields": fields
    }     

    if inputFields != None:
        Body['inputFields'] = Body['inputFields'] + inputFields

    call = restClient.callAPI('/services/apexrest/vlocity_cmt/v2/carts',
                                method="post",
                                data=Body)
    logging.info(f"Quote {name} created with code {restClient.lastCall()['status_code']}")

  #  _set(call['records'][0]['Id'],name,pricelistId,accountId)
    checkError()
    return  call

def createCart_api(accountId,pricelistId,name='', subaction='createOrder',effectivedate =  date.today().strftime("%Y-%m-%d"),channel='python',inputFields = None,fields='Id,Name,EffectiveDate'):
    global _carts

    objectType = "Order"
    if subaction == 'createQuote':
        objectType = "Quote"

    if subaction == 'createOrder':
        Body = {
            "methodName": "createCart",
            "objectType": objectType,
            "inputFields": [
                {"effectivedate": effectivedate },
                {"status": "Draft"},
                {"Name": name},
                {"AccountId": accountId},
                {"vlocity_cmt__PriceListId__c": pricelistId}, 
                {"vlocity_cmt__OriginatingChannel__c": channel}
                ],
            "subaction":subaction,
            "fields": fields
            }

    if inputFields != None:
        Body['inputFields'].append(inputFields)

    call = restClient.callAPI('/services/apexrest/vlocity_cmt/v2/carts',
                                method="post",
                                data=Body)

    checkError()
    logging.info(f"Card {name} created")

   # _set(call['records'][0]['Id'],name,pricelistId,accountId)

    return call

#------------------------------------------------------------------------
def createCart(accountF,pricelistF,name,checkExists=False,subaction='createOrder',effectivedate =  date.today().strftime("%Y-%m-%d"),channel='python',inputFields=None,fields='Id,Name,EffectiveDate'):
    """
    - CartId: 
    """
    if type(accountF) is dict:
        accountId = utils.Id(accountF)
    else:
        accountId = query.queryIdF('product2',accountF)
    
    pricelistId = query.queryIdF('vlocity_cmt__PriceList__c',pricelistF)

    if checkExists:
        cartId = query.queryField(f" select Id from Order where Name ='{name}' and AccountId='{accountId}' ")
        if cartId != None:
            logging.info(f"Cart with Name {name} exists. Returning existing")
            _set(cartId,name,pricelistId,accountId)
            return cartId

    cartCall= createCart_api(accountId,pricelistId,name,subaction,effectivedate,channel,inputFields,fields)

    return cartCall['records'][0]['Id']

def createCartX(accountX,pricelistX,name,checkExists=False,subaction='createOrder',effectivedate =  date.today().strftime("%Y-%m-%d"),channel='python',inputFields=None,fields='Id,Name,EffectiveDate'):
    if type(accountX) is dict:
        accountId = utils.Id(accountX)
    else:
        accountId = query.queryField(f" select Id from Account where $X='{accountX}' ")

    pricelistId = query.queryField(f" select Id from vlocity_cmt__PriceList__c where $X='{pricelistX}' ")

    if checkExists:
        cartId = query.queryField(f" select Id from Order where Name ='{name}' and AccountId='{accountId}' ")
        if cartId != None:
            logging.info(f"Cart with Name {name} exists. Returning existing")
            _set(cartId,name,pricelistId,accountId)
            return cartId

    cartCall= createCart_api(accountId,pricelistId,name,subaction,effectivedate,channel,inputFields,fields)

    return cartCall['records'][0]['Id']

#------------------------------------------------------------------------

def createCartFromAsset_api(orderName, assetId, accountId, date, raiseEx=True):
    if type(assetId) == list:
        assetId = ",".join(assetId)

    data = {
        "subaction": "assetToOrder",
        "id": assetId,
        "accountId": accountId,
        "requestDate": date
    }

    call = restClient.callAPI('/services/apexrest/vlocity_cmt/v2/carts',method="post",data=data)

    call = checkError(call, raiseEx=raiseEx)
    if call == None:
        return None

    id = None
    try:
        id = call['records'][0]['cartId']
    except:
        raise ValueError(str(call['messages']))

    data = {"Name": orderName}
    Sobjects.update(id, data)
    logging.info(f"Created cart {id} from Asset. Order Name <{orderName}>")
    return call

#------------------------------------------------------------------------

def getCartSummary_api(cartid,validate=None,price=None,headerFieldSet=None,translation=None):

    paramStr = stringify(locals(),exclude=['cartid'])
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartid}{paramStr}'

    call = restClient.callAPI(action)
    checkError()
    return call

#------------------------------------------------------------------------

def deleteCart(cartId,cartType='Order'):
    delete = Sobjects.delete('Order',cartId)
    checkError()
    return delete

#------------------------------------------------------------------------

def getCartAttributes_api(cartId):
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/attributes'

    call = restClient.callAPI(action)
    checkError()
    return call

#------------------------------------------------------------------------

def runCartValidation_api(cartid,validate=None,price=None):

    paramStr = stringify(locals(),exclude=['cartid'])
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartid}{paramStr}'

    body=    {
        "cartId":cartid,
        "methodName":"runCartValidation"#,
      #  "price":true,
      #  "validate":true
        }
    call = restClient.callAPI(action,method='post',data=body)
    checkError()
    return call

#-------------------------------------------------------------------------------------------------------------

def getCartItems_api(cartId, query=None,id=None,lastRecordId=None,pagesize=None,hierarchy=None,includeAttachment=None,headerFieldSet=None,filter=None,price=None,validate=None):
    paramStr = stringify(locals(),exclude=['cartid'])

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items{paramStr}'

    call = restClient.callAPI(action)
    checkError()
    return call

#-------------------------------------------------------------------------------------------------------------
def addItemstoCart(cartId,productCode,parentId=None,parentHierarchyPath=None,parentRecord=None,price=None,validate=None,hierarchy=None,pagesize=None,includeAttachment=None,expandAll=None,fields=None,noResponseNeeded=False):
    priceBookEntryId = priceBook.pricebookEntryId_pl(_get(cartId)['plId'],productCode,pricelistField='Id')
    return addItemstoCart_api(cartId,priceBookEntryId,parentId,parentHierarchyPath,parentRecord,price,validate,hierarchy,pagesize,includeAttachment,expandAll,fields,noResponseNeeded)

def addItemstoCart_api(cartId,priceBookEntryId,parentId=None,parentHierarchyPath=None,parentRecord=None,price=None,validate=None,hierarchy=None,pagesize=None,includeAttachment=None,expandAll=None,fields=None,noResponseNeeded=False):
    paramStr = stringify(locals(),exclude=['cartid','itemIds'])

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items{paramStr}'

    data = {
        "items": [{
            "itemId": priceBookEntryId,  
        }]
    }

    if parentId != None:
        data = {
            "items": [{
                "parentId": parentId,
                "parentHierarchyPath": parentHierarchyPath,
                "itemId": priceBookEntryId,  
                "parentRecord": {
                    "records": [
                        parentRecord
                    ]
                }
            }]
        }
    if noResponseNeeded == True:
        data['noResponseNeeded'] = True

    call = restClient.callAPI(action,method='post',data=data)
    logging.info(f"added {priceBookEntryId} to cart {cartId}")
    checkError(call)
    return call

#-------------------------------------------------------------------------------------------------------------

def updateCartItem_api(cartId, items):

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items'
    data = {"items": items}
    call = restClient.callAPI( action, method="put",data=data)
    checkError(call)
    return call

#-------------------------------------------------------------------------------------------------------------

def _getItemAtPath(obj,path,field='ProductCode'):
    for p in path.split(':'):
        obj = objectUtil.getSibling(obj,field,p)  
    return obj

#-------------------------------------------------------------------------------------------------------------

def addToCart_action(parentItem,product,field='ProductCode'):
    sibbling = objectUtil.getSiblingWhere(parentItem,field,product,'itemType','childProduct')['object']
    if sibbling == None:
        logging.warn("The expand does not have a reference to the product. {field}:{product}")
        return None

    actions = objectUtil.getSiblingWhere(sibbling,selectKey='actions')['object']

    if 'records' in parentItem:
        parentItem = parentItem['records'][0]
    parentItem = copy.deepcopy(parentItem)
    parentItem['childProducts'] = ""   

    call = executeActions(actions,'addtocart',parentItem=parentItem)

    logging.info(f"Adding to cart with expansion. Product {product}")

    return call

#-------------------------------------------------------------------------------------------------------------

def addToCart(cartId,parentProductName,productName,field='name',cartItems=None):

    restClient.logCall('getCartItems12')
    if cartItems == None:
        cartItems = getCartItems_api(cartId)

    parentItem = _getItemAtPath(cartItems,parentProductName,field)
    #parentItem = jsonUtil.getSibling(cartItems,'name',parentProductName,logValues=True)
    item = objectUtil.getSibling(parentItem,'name',productName)
   # if item == '':
   #     expanditems = 

    parentItem = copy.deepcopy(parentItem)
    parentItem['childProducts'] = ""

    execute = executeActions(item,parentItem=parentItem)
    return execute

#-------------------------------------------------------------------------------------------------------------

def expand_action(cartItems,sibling,field='ProductCode'):    
    sibblingItem = objectUtil.getSiblingWhere(cartItems,field,sibling,'itemType','lineItem')
    if sibblingItem['object'] == None:
        sibblingItem = objectUtil.getSiblingWhere(cartItems,field,sibling,'itemType','productGroup')

    call = executeActions(sibblingItem['object']['actions'],'expanditems')
    logging.info(f"Expanding {sibling}")
    return call

#-------------------------------------------------------------------------------------------------------------

def expand_api(cartId,itemId,productHierarchyPath):
    paramStr = stringify(locals(),exclude=[], initSeparator='?')

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items/{itemId}/expand{paramStr}'

    call = restClient.callAPI( action, method="get")
    checkError()
    return call

#-------------------------------------------------------------------------------------------------------------

def executeRestAction_api(rest):
    link = rest['link']
    params = rest['params']
    method = f"{rest['method']}".lower()

    call = restClient.callAPI(link, method=method, data=params)
    return call

#-------------------------------------------------------------------------------------------------------------

def executeActions(actionsObject, actionName='addtocart', parentItem=None):
    if 'actions' in actionsObject:
        actions = actionsObject['actions']

    rest = actionsObject[actionName]['rest']
    if actionName == 'addtocart':
        rest['params']['items'][0]['parentRecord'] = {"records": [parentItem]}

    call = executeRestAction_api(rest)
    return call

#-------------------------------------------------------------------------------------------------------------

def deleteCartItems_api(cartId, itemIds,hierarchy=None,lastRecordId=None,pagesize=None,includeAttachment=None,fields=None,query=None):
    paramStr = stringify(locals(),exclude=['cartid','itemIds'],initSeparator='&')

    action =  f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items?id={",".join(itemIds)}{paramStr}'
    requestBody = {
        "items": [{
            "itemId": ",".join(itemIds)
        }]#,
       # "price": True,
       # "validate": True
    }
    call = restClient.callAPI(action,method="delete",data=requestBody)
    checkError()
    return call

#-------------------------------------------------------------------------------------------------------------

def getCartPromotions_api(cartId,getPromotionsAppliedToCart=False,include=None,includePenalties=None,ruleType=None,ruleEvaluationInput=None,filters=None,category=None,fields=None,pagesize=None,commitmentDateFilter=None,appliedPromoStatusFilter=None):
    subaction =  "getPromotionsAppliedToCart" if getPromotionsAppliedToCart == True else None

    paramStr = stringify(locals(),exclude=['cartId','getPromotionsAppliedToCart'])
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/promotions{paramStr}'

    call = restClient.callAPI(action)
    checkError()
    return call['records']

#-------------------------------------------------------------------------------------------------------------

def postCartsPromoItems_api(cartId, promotionId):
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/promotions?cartId={cartId}&id={promotionId}'

    data = {
        "methodName": "postCartsPromoItems",
        "items": [{
            "itemId": promotionId
        }],
        "promotionId": promotionId,
        "cartId": cartId
    }
    call = restClient.callAPI(action, method='post', data=data)
    checkError()

    return call

#-------------------------------------------------------------------------------------------------------------

def deleteCartPromotion_api(cartId, promotionId):
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/promotions?cartId={cartId}&id={promotionId}'
    data = {
        "id": promotionId,
        "cartId": cartId,
        "methodName": "deleteAppliedPromoItems",
        "price": True,
        "validate": True
    }
    call = restClient.callAPI(action, method='delete', data=data)
    checkError()
    return call

#-------------------------------------------------------------------------------------------------------------

def getCartProducts_api(cartId, maxProdListHierarchy=None, query=None, filters=None,getCartProducts=None,lastRecordId=None,includeAttachment=None,offsetSize=None,fields=None,pagesize=None,attributes=None,includeAttributes=None,includeIneligible=True):
    """
    - cartId:  Cart Id (Salesforce Id).Opportunity, Quote or Order Id. - Required. String
    - maxProdListHierarchy: hierarchy depth returned for list of products. Integer. 
    - lastRecordId: The last record ID from the previous search result, if available. string.
    - query: search string. 
    - includeAttachment: Whether product attachments are returned. boolean
    - filters: Filter field values. string
    - offsetSize: Offset from which to start reading products, for pagination. Integet.
    - fields: List of fields to return in the response, separated by commas. 
    - pagesize: Number of records to be returned. integer.
    - attributes: Attribute filters. 
    - includeAttributes: Specifies whether to return a list of attributes and their values for the product. boolean
    """
    paramStr = stringify(locals(),exclude=['cartId'])

    cartId = utils.Id(cartId)
    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/products{paramStr}'

    call = restClient.callAPI(action)

    checkError()
   # if call['messages'][0]['message'] == 'No Results Found.':
   #     return None
    return call['records']

#-------------------------------------------------------------------------------------------------------------

def checkOut_api(cartId,validateSubmittedXLI=True,assetizeFullBundlePerRoot=True,checkOrderStatus=True, provisioningStatus='Active',skipCheckoutValidation=True,waitActivation=False,asyncCall = False):
    cartId = utils.Id(cartId)

    data = {
        "methodName": "checkout",
        "cartId": cartId,
        "ContextId": cartId,
        #     "objectTypeToIdsToClone":"OrderItem:8021N000008NG5t_8021N000008NG5s",
        "validateSubmittedXLI": validateSubmittedXLI,
        "assetizeFullBundlePerRoot": assetizeFullBundlePerRoot,
        "checkOrderStatus": checkOrderStatus,
        "provisioningStatus": provisioningStatus,
        "skipCheckoutValidation": skipCheckoutValidation
    }

    action = f'/services/apexrest/vlocity_cmt/v2/cpq/carts/{cartId}/items/checkout'
    if asyncCall == True:
        action = f'/services/apexrest/AsyncCPQAppHandler/v1/checkout'
    call = restClient.callAPI(
        action,
        method="post",
        data=data)
    logging.info(f'Cart Check-out with code {call["status_code"]}')
    checkError()
    assert(False)
   # if waitActivation == True:
   #   orderUtils.waitforOrderActivation(cartId)
    
    return call



import json
from .SFAPI import Sobjects, query,restClient,jsonFile,utils,objectUtil,debugLogs,digitalCommerceUtil
import logging
import simplejson,sys
from operator import itemgetter
from inspect import getmembers, isfunction
import traceback


varsFile = 'confIcli.json'
vars = None


def readVars(field=None):
    global vars
    vars = jsonFile.read(varsFile)

    if field != None:
        return vars[field]
    return vars

def setVar(name,value):

    vars = jsonFile.read(varsFile)
    vars[name]=value
    jsonFile.write(varsFile,vars)
    print(simplejson.dumps(vars, indent=4))

def getParam(argv,key,i=1):
    if key not in argv:
        return None

    ind = argv.index(key)
    if (ind+i <= (len(argv)-1)):
        arg = argv[ind+i]
        if arg[0] == '-':
            return None
        return arg
    return None

def listEnvs():
    cons = restClient.getConfigOrgs()
    cons = sorted(cons,key=itemgetter('name'))
    utils.printFormated(cons,fieldsString='name:isSandBox:instance_url:login.username:login.password:login.bearer')
    env = readVars('environment')
    print(f"Current Environment is {env}")

def _queryAndPrint(q,fields=None,systemFields=True,nullFields=False):

    res = query.queryRecords(q)
    if fields != None:
        res = utils.deleteNulls(res,systemFields,nullFields)
        if fields=='all':
            utils.printFormated(res)
        else:
            utils.printFormated(res,fields)
    else:
        res = utils.deleteNulls(res,systemFields,nullFields)
        print(simplejson.dumps(res, indent=4))
    
    print()
    print(f"Null values printed-> {nullFields}  systemFields printed--> {systemFields}")

def option_q(args):
    if '-h' in args:
        help = f"""
        {utils.CGREEN}-q "select..."{utils.CEND} query to execute. System Fields, such as LastModifiedDate, and null fields are not returned. 
            {utils.CGREEN}-null{utils.CEND} - will print fields with null values
            {utils.CGREEN}-system{utils.CEND} - will print the system fields
            {utils.CGREEN}-all{utils.CEND} - will print both nulls and system fields
            {utils.CGREEN}-fields "a:b:c:..."{utils.CEND} --> print in table format the specified fields.
            {utils.CGREEN}-fields all{utils.CEND} --> print in table format"""
        return help

    #else:
    q = getParam(args,'-q')

    fields = getParam(args,'-fields') if '-fields' in args else None    
    nullFields = True if '-null' in args else False
    systemFields = True if '-system' in args else False
    all = True if '-all' in args else False
    if all == True:
        nullFields = True
        systemFields = True

    connectionInit(args)

    _queryAndPrint(q,fields=fields,systemFields=systemFields,nullFields=nullFields)

def option_cc(args):
    if '-h' in args:
        help = f"""
        {utils.CGREEN}-cc{utils.CEND} 
            Checks the catalogs. Gets all catalogs, does a getOffers, getOfferDetails, basketwithoutconfig and basket with config
            {utils.CGREEN}-code{utils.CEND} - Catalogue Code. does it for a single catalog. 
            {utils.CGREEN}-list{utils.CEND} - List all the catalogues. 
            """ 
        return help

    connectionInit(args)

   # path = getParam(args,'-checkCatalogs')
   # quantity = getParam(args,'-checkCatalogs',2)
    catcatalogueCode = getParam(args,'-code')
    list = True if '-list' in args else False
    account = getParam(args,'-account')

    if list == True:
        digitalCommerceUtil.printCatalogs()
        return

    digitalCommerceUtil.checkOffers(catcatalogueCode=catcatalogueCode,account=account)

    print()

def option_d(args):
    help = f"""
        {utils.CGREEN}-d objectName{utils.CEND} --> Describe an object
            {utils.CGREEN}-d objectName:fieldName{utils.CEND} --> describe a field in the objec"""
    if '-h' in args:
        return help    

    connectionInit(args)

    objectField = getParam(args,'-d')
    if objectField == None:
        print(help)
        return

    ofs = objectField.split(':')

    sObjectName = ofs[0]
    fieldName = ofs[1] if len(ofs) > 1 else None

    res = Sobjects.describe(sObjectName)
    if fieldName == None:
        print(simplejson.dumps(res['fields'], indent=4))
    else:
        sibbling = objectUtil.getSiblingWhere(res['fields'],'name',fieldName)['object']

        print(simplejson.dumps(sibbling, indent=4))

def option_l(args):
    if '-h' in args:
        help = f"""
        {utils.CGREEN}-l{utils.CEND} --> current org limits consuptions. """
        return help
    connectionInit(args)
    action = '/services/data/v51.0/limits'
    res = restClient.callAPI(action)
    records = []
    for key in res.keys():
        record = {
            'Limit':key,
            'Max':res[key]['Max'],
            'Remaining':res[key]['Remaining'],
        }
        record['Percent Remaining'] =  100 *(res[key]['Remaining']/res[key]['Max']) if res[key]['Max']>0 else 0
        record['__color__'] = ''

        if record['Max'] != 0:
            if record['Percent Remaining']<50:
                record['__color__'] = utils.CYELLOW
            if record['Percent Remaining']<25:
                record['__color__'] = utils.CYELLOW
        records.append(record)

    utils.printFormated(records)

def option_o(args):
    if '-h' in args:
        help = f"""
        {utils.CGREEN}-o{utils.CEND} --> List all Objects
            {utils.CGREEN}-name objectName{utils.CEND}  --> get one row from the object
            {utils.CGREEN}-name objectName:Id{utils.CEND} --> get the row especified by the Id
            {utils.CGREEN}-like name{utils.CEND} --> where the Name contains the substring name"""

        return help
        
    connectionInit(args)

    if '-name' in args:
        obj = getParam(args,'-name')
        chunks = obj.split(':')

        if len(chunks) == 1:
            q = f"select fields(all) from {chunks[0]} limit 1"
        else:
            q = f"select fields(all) from {chunks[0]} where Id='{chunks[1]}'"

        print('pre q:'+q)
        _queryAndPrint(q,systemFields=True,nullFields=True)
        return
        
    like = getParam(args,'-like') if '-like' in args else None    
    count = True if '-count' in args else False   

    objs = Sobjects.listObjects()
    if like is not None:
        if ':' not in like:
            like = f"name:{like}"        
        ls = like.split(':')
        outs = []
        for obj in objs:
      #      print(f"{str(ls[1])} . {str(obj[ls[0]])}")
            if str(ls[1]).lower() in str(obj[ls[0]]).lower():
                outs.append(obj) 
    else:
        outs = objs

    if count==True:
        for out in outs:
            if out['queryable'] == True:
                print("Quering objects row count.")
                print("", end=".")          
                try:
                    if out['name'] == 'AccountUserTerritory2View':
                        print()
                    c = query.query(f" select count(Id) from {out['name']}",raiseEx=True)
                    out['count'] = c['records'][0]['expr0']
                except Exception as e:
                    out['count'] = 'E'
            else:
                out['count'] = '-'
    
    #print(simplejson.dumps(objs, indent=4))
    utils.printFormated(outs,'name:label:associateParentEntity:associateParentEntity:queryable:count')

def option_default(args):
    help = f"""
        {utils.CGREEN}-default:set key value{utils.CEND} --> defaults the specified key-value
        {utils.CGREEN}-default:del key  {utils.CEND}  --> deletes the default
        {utils.CGREEN}-default:get key  {utils.CEND}  --> displays the current value
        """

    if '-h' in args:
        return help   

    if "-default:del" in args:
        key = getParam(args,'-default:del',1)
        if key == None:
            restClient.glog().info(f'Key not found in the provided arguments. {args}')
            return 
        restClient.delConfigVar(key)
        return

    if "-default:set" in args: 
        key = getParam(args,'-default:set',1)
        value = getParam(args,'-default:set',2)
        if key == None or value==None:
            print(help)
            return

        restClient.setConfigVar(key,value)
        print(f"Default value for {key} is {value}")

    if "-default:get" in args: 
        key = getParam(args,'-default:get',1)
        value =restClient.getConfigVar(key)
        print(f"Default value for {key} is {value}")

#def option_history(args):
#    return
#    if '-h' in args:
        help = """
        -history"""
        return help
#    print()
#    print('HISTORY:')
#    vars = jsonFile.read(varsFile)
#    for line in vars['history']:
        print(line)

def option_logs(args):
    help = f"""
        {utils.CGREEN}-logs{utils.CEND} --> list logs in the org. 
            {utils.CGREEN}-limit X{utils.CEND}, where X specifies the number of logs to list. Default 50 max 50K
            {utils.CGREEN}-loguser field:value{utils.CEND}, filter the logs for the specified user. The user can be specified by any field in the User Object. 
                Examples:
                    "Id:XXXXXXXX"
                    "name:Onboarding Site Guest User"
                    "Alias:John.Doe"
                    "FirstName:Onboarding"
                    "ProfileId:00e3O000000IHneQAG"
                {utils.CPURPLE}Defaultable loguser{utils.CEND}: Can be set to default. -default:set loguser "Alias:John.Doe"
        {utils.CGREEN}-logs Id{utils.CEND} --> parse the logs with the provided Id
        {utils.CGREEN}-logs -last X {utils.CEND}--> parses the last X logs. 
        """

    if '-h' in args:
        return help

    logId = getParam(args,'-logs')    
    last = getParam(args,'-last')  
    limit = getParam(args,'-limit')    
    level = getParam(args,'-level') 
    loguser = getParam(args,'-loguser') 
    whereClause = getParam(args,'-where') 

    if loguser == None:
        loguser = restClient.getConfigVar('loguser')

    if logId == None and last== None:
        connectionInit(args)
        lim = 50 if limit == None else limit
        id = None
        if loguser != None:
            chunks = loguser.split(":")
            key = chunks[0] if len(chunks)>1 else 'Id'
            value = chunks[1] if len(chunks)>1 else chunks[0]
            id = query.queryField(f"Select Id from User where {key}='{value}'") if key.lower()!='id' else value
            if id == None:
                utils.raiseException('QUERY',f"User with field {key} = {value} does not exist in the User Object.")
        if id != None:
            print(f"Logs for user {loguser}, id:{id}")
        debugLogs.printLogs(logUserId=id,limit=lim,whereClause=whereClause)
        return

    connectionInit(args)
    debugLogs.parseLog(logId,lastN=last,level=level,whereClause=whereClause)

    return

def option_h(args):
    module = __import__('InCli')

    funcs = getmembers(sys.modules[__name__], isfunction)
    functions = [func[0] for func in funcs if func[0].startswith('option_')]

    for f in functions:
        if f == 'option_h':
            continue

        print(eval(f'{f}(args)'))      

def connectionInit(argsIn):

    userName_or_ofgAlias = getParam(argsIn,'-u') 
    restClient.init(userName_or_ofgAlias)

def _main(argsIn):
    restClient.setLoggingLevel(logging.INFO)

    if '-h' in argsIn:
        help = f"""
    {utils.CGREEN}-u{utils.CEND} --> username or org alias to be used to log into the org. 
            {utils.CPURPLE}Defaultable u{utils.CEND}: Can be set to a default value --> -default:set u "xx@adas.com"
    """
        print(help)

    funcs = getmembers(sys.modules[__name__], isfunction)
    functions = [func[0] for func in funcs if func[0].startswith('option_')]

    args = []
    for argv in argsIn:
        if argv == '|':
            break
        args.append(argv)

    for arg in args:
        ar = arg.split(':')[0][1:]
        ar = f"option_{ar}"
        if ar in functions:
        #       module = __import__('InCli')
            eval(f'{ar}(args)')
        #      func = getattr(module, ar)
        #      func(args)

    if '-h' in argsIn:
        print()
        print(utils.CBOLD+"SFDX Commands:"+utils.CEND)
        print(f" - {utils.CGREEN}sfdx force:org:list --verbose --all {utils.CEND}--> to list all authorized Orgs and Connection Status")
        print(f" - {utils.CGREEN}sfdx auth:web:login -r 'Instance Url' -a 'Alias' {utils.CEND}--> to re-authorize")
        print(f" - {utils.CGREEN}sfdx auth:web:login -u 'userName'  -a 'alias' {utils.CEND}--> to authorize and Org")
        print()

def main():
    argsIn = sys.argv
    try:
        _main(argsIn)

    except Exception as e:
        utils.printException(e)
        traceback.format_exc()

if __name__ == '__main__':
    main()

